#External:
import os
import shutil

import ftplib
import copy
import numpy as np

unique_file_id_list=['checksum_type','checksum','tracking_id']

def download_secure_FTP(url_name,dest_name,username=None,user_pass=None):
    if (username!=None and 
        user_pass!=None):
        #Use credentials:
        ftp=ftplib.FTP(url_name.split('/')[2],username,user_pass)

    else:
        #Do not use credentials and hope for anonymous:
        ftp=ftplib.FTP(url_name.split('/')[2])

    directory=os.path.dirname(dest_name)
    if not os.path.exists(directory):
        os.makedirs(directory)

    with open(dest_name,'wb') as local_file:
        try:
            ftp.retrbinary('RETR %s' % '/'+'/'.join(url_name.split('/')[3:]), local_file.write)
        except ftplib.error_perm:
            #Permission error. Try again!
            ftp.retrbinary('RETR %s' % '/'+'/'.join(url_name.split('/')[3:]), local_file.write)
    
    ftp.close()
    file_size=np.float(os.stat(dest_name).st_size)
    return "Downloading: %s MB: %s" % (dest_name, file_size/2.0**20)

def download_secure_local(url_name,dest_name):
    directory=os.path.dirname(dest_name)
    if not os.path.exists(directory):
        os.makedirs(directory)
    shutil.copy(url_name,dest_name)
    file_size=np.float(os.stat(dest_name).st_size)
    return "Downloading: %s MB: %s" % (dest_name, file_size/2.0**20)





#External:
import argparse 
import textwrap
import datetime
import copy

def input_arguments(parser):
    parser.add_argument('in_netcdf_file',
                     help='NETCDF paths file (input)')
    return

def output_arguments(parser):
    parser.add_argument('out_netcdf_file',
                     help='NETCDF retrieved file (output)')
    return

def generate_subparsers(project_drs):
    #Option parser
    version_num='0.7.8.1'
    description=textwrap.dedent('''\
    This script aggregates soft links to OPENDAP or local files.\
    ''')
    epilog='Version {0}: Frederic Laliberte (09/2016),\n\
Previous versions: Frederic Laliberte, Paul Kushner (2011-2016).\n\
\n\
If using this code to retrieve and process data from the ESGF please cite:\n\n\
Efficient, robust and timely analysis of Earth System Models: a database-query approach (2016):\n\
F. Laliberte, Juckes, M., Denvil, S., Kushner, P. J., TBD, Submitted.'.format(version_num)
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                            description=description,
                            version='%(prog)s '+version_num,
                            epilog=epilog)


    subparsers = parser.add_subparsers(help='Commands to organize and retrieve data from heterogenous sources',dest='command')
    certificates(subparsers,epilog,project_drs)
    subset(subparsers,epilog,project_drs)

    validate(subparsers,epilog,project_drs)
    download_files(subparsers,epilog,project_drs)
    download_opendap(subparsers,epilog,project_drs)
    load(subparsers,epilog,project_drs)
    return parser

def subset(subparsers,epilog,project_drs):
    epilog_validate=textwrap.dedent(epilog)
    parser=subparsers.add_parser('subset',
               description=textwrap.dedent('Returns a\n\
                     netcdf file subsetted along latitudes and longitudes.'),
               epilog=epilog_validate,
             )
    parser.add_argument('--lonlatbox',default=[0.0,359.999,-90.0,90.0],
                      nargs=4,type=float,
                     help='Longitude - Latitude box in degrees. Default: 0.0 359.999 -90.0 90.0')
    parser.add_argument('--lat_var',default='lat',
                     help='Name of latitude variable')
    parser.add_argument('--lon_var',default='lon',
                     help='Name of longitude variable')
    parser.add_argument('--output_vertices',action='store_true',help='Compute and output vertices')
    input_arguments(parser)
    output_arguments(parser)
    return

def validate(subparsers,epilog,project_drs):
    epilog_validate=textwrap.dedent(epilog)
    parser=subparsers.add_parser('validate',
               description=textwrap.dedent('Returns a\n\
                     netcdf file with soft links to the data. Validates availability of remote data.'),
               epilog=epilog_validate,
             )
    validate_arguments(parser,project_drs)
    certificates_arguments(parser,project_drs)
    return

def validate_arguments(parser,project_drs):
    parser.add_argument('var_name',
                     default=[], type=str_list,
                     help='Comma-seprated variable names to concatenate.')
    parser.add_argument('in_netcdf_file',nargs='*',
                     help='NETCDF paths file (input)')
    output_arguments(parser)

    time_selection_arguments(parser,project_drs)
    parser.add_argument('--file_type',default='local_file',choices=['local_file','OPENDAP'],
                     help='Type of files.')
    parser.add_argument('--time_var',default='time',
                     help='The time variable in the files. Default: time.')

    return

def download_files(subparsers,epilog,project_drs):
    epilog_download_files=textwrap.dedent(epilog)
    parser=subparsers.add_parser('download_files',
                   description=textwrap.dedent('Take as an input the results from \'validate\' and returns a\n\
                                                soft links file with the HTTPServer / FTPServer and GRIDFTP data filling the database.'),
                   epilog=epilog_download_files,
                 )
    download_files_arguments(parser,project_drs)
    parser.add_argument('--time_var',default='time',help='Name of time variable. Default=time.')
    return parser

def download_files_arguments(parser,project_drs):
    input_arguments(parser)
    output_arguments(parser)
    download_files_arguments_no_io(parser,project_drs)
    download_arguments_no_io(parser,project_drs)
    serial_arguments(parser,project_drs)
    certificates_arguments(parser,project_drs)
    data_node_restriction(parser,project_drs)
    time_selection_arguments(parser,project_drs)
    return parser

def download_opendap(subparsers,epilog,project_drs):
    epilog_validate=textwrap.dedent(epilog)
    parser=subparsers.add_parser('download_opendap',
               description=textwrap.dedent('Take as an input the results from \'validate\' and returns a\n\
                                            soft links file with the opendap data filling the database.\n\
                                            Must be called after \'download_files\' in order to prevent missing data.'),
               epilog=epilog_validate,
             )
    download_opendap_arguments(parser,project_drs)
    parser.add_argument('--time_var',default='time',help='Name of time variable. Default=time.')
    return parser

def download_opendap_arguments(parser,project_drs):
    input_arguments(parser)
    output_arguments(parser)
    download_opendap_arguments_no_io(parser,project_drs)
    download_arguments_no_io(parser,project_drs)
    serial_arguments(parser,project_drs)
    certificates_arguments(parser,project_drs)
    data_node_restriction(parser,project_drs)
    time_selection_arguments(parser,project_drs)
    return parser

def download_opendap_arguments_no_io(parser,project_drs):
    parser.add_argument('--download_all_opendap',default=False,action='store_true',
                        help='Download all remote opendap links, even if a local_file is available.\n\
                              By default, download only OPENDAP links that have no alternatives.')
    return

def download_files_arguments_no_io(parser,project_drs):
    if project_drs==None:
        default_dir='.'
    else:
        default_dir='./'+project_drs.project
    parser.add_argument('--out_download_dir',default='.',
                             help='Destination directory for retrieval.')
    parser.add_argument('--download_all_files',default=False,action='store_true',
                        help='Download all remote files corresponding to the request, even if a local_file or an OPENDAP link are available.\n\
                              By default, download only files that have no alternatives.')

    parser.add_argument('--do_not_revalidate',default=False,action='store_true',
                        help='Do not revalidate. Only advanced users will use this option.\n\
                              Using this option might can lead to ill-defined time axes.')
    return 

def download_arguments_no_io(parser,project_drs):
    parser.add_argument('--download_cache',help='Cache file for downloads')
    parser.add_argument('--num_dl',default=1,type=int,help='Number of simultaneous download from EACH data node. Default=1.')
    return parser

def serial_arguments(parser,project_drs):
    parser.add_argument('--serial',default=False,action='store_true',help='Force serial analysis.')
    return parser

def load(subparsers,epilog,project_drs):
    epilog_load=textwrap.dedent(epilog)
    parser=subparsers.add_parser('load',
                   description=textwrap.dedent('Take as an input the results from \'validate\' and loads local data into the database.\n\
                                                Removes soft links informations. Must be used after download_files and download_opendap in order\n\
                                                to prevent missing data.'),
                   epilog=epilog_load,
                 )
    input_arguments(parser)
    output_arguments(parser)
    verbosity_group = parser.add_argument_group('Specify verbosity in downloads')
    verbosity_group.add_argument('-s','--silent',default=False,action='store_true',help='Make downloads silent.')
    certificates_arguments(parser,project_drs)
    time_selection_arguments(parser,project_drs)
    return parser

def certificates(subparsers,epilog,project_drs):
    epilog_certificates=textwrap.dedent(epilog)
    parser=subparsers.add_parser('certificates',
                           description=textwrap.dedent('Recovers ESGF certificates. This method of obtaining credentials \n\
                                                        has been superseded by the more robust cookie method in netcdf4_soft_links version 0.6 \n\
                                                        Will most likely become deprecated in the near future. '),
                           epilog=epilog_certificates
                         )
    certificates_arguments(parser,project_drs)
    return

def time_selection_arguments(parser,project_drs):
    time_inc_group = parser.add_argument_group('Time selection')
    time_inc_group.add_argument('--year',
                                 default=None, type=int_list,
                                 help='Retrieve only these comma-separated years.')
    time_inc_group.add_argument('--month',
                                 default=None, type=int_list,
                                 help='Retrieve only these comma-separated months (1 to 12).')
    time_inc_group.add_argument('--day',
                                 default=None, type=int_list,
                                 help='Retrieve only these comma-separated calendar days.')
    time_inc_group.add_argument('--hour',
                                 default=None, type=int_list,
                                 help='Retrieve only these comma-separated hours.')
    time_inc_group.add_argument('--previous',
                                 default=0,action='count',
                                 help='Retrieve data from specified year, month, day AND the time step just BEFORE this retrieved data.')
    time_inc_group.add_argument('--next',
                                 default=0,action='count',
                                 help='Retrieve data from specified year, month, day AND the time step just AFTER this retrieved data.')
    return

def certificates_arguments(parser,project_drs):
    if project_drs==None:
        script_name='nc_soft_links'
    else:
        script_name='cdb_query '+project_drs.project

    cert_group = parser.add_argument_group('Use these arguments to let cdb_query manage your credentials')
    cert_group.add_argument('--openid',default=None,
                     help='Pass your ESGF openid.\n\
                           cdb_query will prompt you once for your password and will ensure that your credentials \n\
                           are active for the duration of the process. If your openid starts with https://ceda.ac.uk, you\n\
                           must, in addition, pass a username.')
    cert_group.add_argument('--username',default=None,
                            help='Pass you username. Necessary for compatibility with CEDA openids or for \n\
                             FTP queries.')
    cert_group.add_argument('--password', help='Your ESGF password')
    cert_group.add_argument('--password_from_pipe',default=False,action='store_true',
                        help='If activated it is expected that the user is passing a password through piping\n\
                              Example: echo $PASSWORD | '+script_name+' ...')
    cert_group.add_argument('--timeout',default=120,type=int,
                     help='Set the time after which opendap access will timeout (in seconds).\n\
                           If a connection is slow, TIMEOUT should probably be larger. Default: 120s.')

    cert_group.add_argument('--use_certificates',default=False,action='store_true',
                     help='Use certificates. Only valid with a CEDA openid (e.g. https://ceda.ac.uk/openid/OPENID). \n\
                           Will likely be deprecated in future versions.')
    cert_group.add_argument('--no_trustroots',default=True,action='store_false',
                     help='Bypass trustroots retrieval. Retrieval can be slow and is necessary only one in a while. \n\
                           Will liekly be deprecated in future versions.')
    return

def data_node_restriction(parser,project_drs):
    data_node_group = parser.add_argument_group('Restrict to specific data nodes')
    data_node_group.add_argument('--data_node',type=str,action='append',help='Consider only the specified data nodes')
    data_node_group.add_argument('--Xdata_node',type=str,action='append',help='Do not consider the specified data nodes')
    return

def int_list(input):
    return [ int(item) for item in input.split(',')]

def str_list(input):
    if len(input.split(','))==1:
        return input
    else:
        return [ item for item in input.split(',')]



#External:
import os, sys, string, subprocess
import getpass
import select
import pydap.lib

#External by related:
from netcdf4_pydap.cas.esgf import get_node

#Internal:
from .onlineca_get_trustroots_wget import onlineca_get_trustroots_wget
from .onlineca_get_cert_wget import onlineca_get_cert_wget

def prompt_for_username_and_password(options):
    if (options.command=='certificates' and 
        'openid' in dir(options) and options.openid==None):
        options.openid=raw_input('Enter OpenID: ')

    if (('openid' in dir(options) and options.openid!=None) and
        get_node(options.openid)=='https://ceda.ac.uk'):
        if (options.command=='certificates' and 
            'username' in dir(options) and options.username==None):
            options.username=raw_input('Enter CEDA Username: ')
    else:
        if options.command=='certificates':
            raise InputError('Only OpenIDs from CEDA (starting with https://ceda.ac.uk) can \n\
                              be used to retrieve certificates.')
        elif ('use_certificates' in dir(options) and options.use_certificates):
            raise InputError('Only OpenIDs from CEDA (starting with https://ceda.ac.uk) can \n\
                              be used to retrieve certificates. Do not use --use_certificates.')
        
    if not 'password' in dir(options):
        options.password=None
    elif ( 'openid' in dir(options) and 
         options.openid!=None and
         options.password == None):

        if not options.password_from_pipe:
            options.password = getpass.getpass('Enter Credential phrase: ').strip()
        else:
            prompt_timeout=1
            i,o,e=select.select([sys.stdin],[],[],prompt_timeout)
            if i:
                options.password = sys.stdin.readline().strip()
            else:
                print('--password_from_pipe selected but no password was piped. Exiting.')
                return

    #Retrieve certificates or set dods_conf:
    if (('use_certificates' in dir(options) and options.use_certificates) 
         or  options.command=='certificates'):
        registering_service='ceda'
        if 'username' in dir(options) and options.username!=None:
            retrieve_certificates(options.username, registering_service, user_pass=options.password,
                                  trustroots=options.no_trustroots,
                                  timeout=options.timeout)
        else:
            retrieve_certificates(None, registering_service)
    return options

def retrieve_certificates(username,registering_service,user_pass=None,trustroots=False,timeout=120):
    home=os.getenv('HOME')

    dods=True
    dodsrc='%s/.dodsrc' % home
    esgfdir='%s/.esg4' % home
    if username==None and dods:
      #Set the environment:
      os.environ['DODS_CONF']=dodsrc
      #This modifies pydap. All three lines must be loaded in order:
      #pydap.lib.CACHE=esgfdir+'/dods_cache'
      #import pydap_esgf
      #pydap_esgf.install_esgf_client(os.environ['X509_USER_PROXY'],os.environ['X509_USER_PROXY'])
      return

    http_proxy=os.getenv('http_proxy')
    https_proxy=os.getenv('https_proxy')
    if http_proxy != None and https_proxy== None:
      print('You have http_proxy set but not https_proxy: download tests are likely to fail')

    ee = { 'smhi':'esg-dn1.nsc.liu.se', 'pcmdi':'pcmdi9.llnl.gov', 'ipsl':'esgf-node.ipsl.fr', 'ceda':'slcs.ceda.ac.uk',
    'dkrz':'esgf-data.dkrz.de', 'pik':'esg.pik-potsdam.de', 'jpl':'jpl-esg.jpl.nasa.gov' }

    registering_service = ee.get( registering_service, registering_service )

    for f in [esgfdir,esgfdir+'/certificates',esgfdir+'/dods_cache']:
      if not os.path.isdir( f ):
        os.mkdir( f )

    dodstext = """## generated by enesGetCert
HTTP.COOKIEJAR=%(esgfdir)s/dods_cache/.dods_cookies
HTTP.SSL.CERTIFICATE=%(esgfdir)s/credentials.pem
HTTP.SSL.KEY=%(esgfdir)s/credentials.pem
HTTP.SSL.CAPATH=%(esgfdir)s/certificates
HTTP.TIMEOUT=%(timeout)s"""
#CURL.VERBOSE=0
#CURL.COOKIEJAR=.dods_cookies
#CURL.SSL.CERTIFICATE=%(esgfdir)s/credentials.pem
#CURL.SSL.KEY=%(esgfdir)s/credentials.pem
#CURL.SSL.CAPATH=%(esgfdir)s/certificates

    if dods:
      oo = open( dodsrc, 'w' )
      oo.write( dodstext % locals() )
      oo.close()
      #Temporary fix: write a local .dodsrc
      #oo = open( '.dodsrc', 'w' )
      #oo.write( dodstext % locals() )
      #oo.close()


    oo = open(esgfdir+'/onlineca-get-trustroots-wget.sh','w')
    oo.write(onlineca_get_trustroots_wget())
    oo.close()

    if trustroots:
        #Do not have to renew frequently
        call_to_script=['bash',esgfdir+'/onlineca-get-trustroots-wget.sh','-b','-U','https://'+registering_service+'/onlineca/trustroots/','-c',esgfdir+'/certificates']
        #print ' '.join(call_to_script)
        subprocess.call(call_to_script)

    oo = open(esgfdir+'/onlineca-cert-wget.sh','w')
    oo.write(onlineca_get_cert_wget())
    oo.close()
    call_to_script=['bash',esgfdir+'/onlineca-cert-wget.sh','-l',username,'-U','https://'+registering_service+'/onlineca/certificate/','-c',esgfdir]
    #print ' '.join(call_to_script)
    if user_pass!=None:
        call_to_script.append('-S')
        p=subprocess.Popen(call_to_script,stdout=subprocess.PIPE,stdin=subprocess.PIPE)
        stdout=p.communicate(input=user_pass)[0]
        p.stdin.close()
    else:
        subprocess.call(call_to_script)

    #Remove cookies for a clean slate:
    try:
        os.remove(esgfdir+'/dods_cache/.dods_cookies')
    except:
        pass
    return



#External:
import numpy as np
import tempfile
import copy
import os
import datetime

#Internal:
from ..remote_netcdf import remote_netcdf
from .. import netcdf_utils

queryable_file_types = ['OPENDAP', 'local_file', 'soft_link_container']
unique_file_id_list = ['checksum_type', 'checksum', 'tracking_id']

class create_netCDF_pointers:
    def __init__(self, paths_list, time_frequency, years, months,
                 file_type_list, data_node_list,
                 semaphores=dict(), record_other_vars=True, check_dimensions=False,
                 time_var='time', session=None, remote_netcdf_kwargs=dict()):
        self.semaphores = semaphores
        self.session = session
        self.remote_netcdf_kwargs = remote_netcdf_kwargs
        self.paths_list = paths_list

        if check_dimensions:
            #Use this option to ensure some sort of dimensional compatibility:
            sorts_list = ['version', 'file_type_id', 'dimension_type_id', 'data_node_id', 'path_id']
        else:
            sorts_list = ['version', 'file_type_id', 'data_node_id', 'path_id']
        self.id_list = ['data_node', 'file_type', 'path']+unique_file_id_list

        self.time_frequency = time_frequency
        self.is_instant = False
        self.record_other_vars = record_other_vars
        self.time_var = time_var

        self.months = months
        self.years = years

        self.paths_ordering = order_paths_by_preference(sorts_list,
                                                      self.id_list,
                                                      self.paths_list,
                                                      file_type_list,
                                                      data_node_list,
                                                      check_dimensions,
                                                      semaphores=self.semaphores,
                                                      time_var=self.time_var,
                                                      session=self.session,
                                                      remote_netcdf_kwargs=self.remote_netcdf_kwargs)
        return

    def record_paths(self,output,var):
        return self.create(output)

    def record_meta_data(self,output,var):
        if self.time_frequency in ['fx','clim']:
            if isinstance(var,list):
                for sub_var in var:
                    self.record_fx(output,sub_var)
            else:
                self.record_fx(output,var)
        else:
            self.calendar = obtain_unique_calendar(self.paths_ordering,semaphores=self.semaphores,
                                                                     time_var=self.time_var,
                                                                     session=self.session,
                                                                     remote_netcdf_kwargs=self.remote_netcdf_kwargs)
            #Retrieve time and meta:
            self.create_variable(output,var)
            #Put version:
        netcdf_utils.setncattr(output,str('netcdf_soft_links_version'),str('1.3'))
        return

    def record_fx(self,output,var):
        #Find the most recent version:
        most_recent_version = 'v'+str(np.max([int(item['version'][1:]) for item in self.paths_list]))
        usable_paths_list = [ item for item in self.paths_list if item['version']==most_recent_version]

        queryable_paths_list = [item for item in usable_paths_list if item['file_type'] in queryable_file_types]
        if len(queryable_paths_list)==0:
            temp_file_handle, temp_file_name = tempfile.mkstemp()
        try:
            if len(queryable_paths_list)==0:
                path = usable_paths_list[0]
                #Download the file to temp
                #retrieval_utils.download_secure(path['path'].split('|')[0],
                #                temp_file_name,
                #                path['file_type'],
                #                username=username,user_pass=user_pass)
                #remote_data=remote_netcdf.remote_netCDF(temp_file_name,path['file_type'],self.semaphores,**self.remote_netcdf_kwargs)
            else:
                #Check if data in available:
                path = queryable_paths_list[0]

                remote_data = remote_netcdf.remote_netCDF(path['path'].split('|')[0],path['file_type'],semaphores=self.semaphores,
                                                                                                     session=self.session,
                                                                                                     **self.remote_netcdf_kwargs)
                alt_path_name = remote_data.check_if_available_and_find_alternative([item['path'].split('|')[0] for item in queryable_paths_list],
                                                                            [item['file_type'] for item in queryable_paths_list],
                                                                         [item['path'].split('|')[unique_file_id_list.index('checksum')+1] for item in queryable_paths_list],
                                                                         queryable_file_types)

                #Use aternative path:
                path = queryable_paths_list[[item['path'].split('|')[0] for item in queryable_paths_list].index(alt_path_name)]
                remote_data = remote_netcdf.remote_netCDF(path['path'].split('|')[0],path['file_type'],semaphores=self.semaphores,
                                                                                                     session=self.session,
                                                                                                     **self.remote_netcdf_kwargs)

            output = remote_data.safe_handling(netcdf_utils.retrieve_variables,output,zlib=True)

            for att in path.keys():
                if att!='path':      
                    netcdf_utils.setncattr(output,att,path[att])
            netcdf_utils.setncattr(output,'path',path['path'].split('|')[0])
            for unique_file_id in unique_file_id_list:
                netcdf_utils.setncattr(output,unique_file_id,path['path'].split('|')[unique_file_id_list.index(unique_file_id)+1])
        finally:
            if len(queryable_paths_list)==0:
                os.remove(temp_file_name)

        #Create soft links
        output_grp = self.create(output)
        output_grp.createVariable(var,np.float32,(),zlib=True)
        return

    def create(self,output):
        if not 'soft_links' in output.groups.keys():
            output_grp = output.createGroup('soft_links')
        else:
            output_grp = output.groups['soft_links']

        #OUTPUT TO NETCDF FILE PATHS DESCRIPTIONS:
        output_grp.createDimension('path',None)
        for id in ['version','path_id']:
            output_grp.createVariable(id,np.int64,('path',),chunksizes=(1,),zlib=True)[:]=self.paths_ordering[id]
        for id in self.id_list:
            temp = output_grp.createVariable(id,str,('path',),chunksizes=(1,),zlib=True)
            for file_id, file in enumerate(self.paths_ordering['path']):
                temp[file_id] = str(self.paths_ordering[id][file_id])
        return output_grp

    def create_variable(self,output,var):
        #Recover time axis for all files:
        date_axis, table, units = obtain_date_axis(self.paths_ordering,
                                                self.time_frequency,
                                                self.is_instant,
                                                self.calendar,
                                                semaphores=self.semaphores,
                                                time_var=self.time_var,
                                                session=self.session,
                                                remote_netcdf_kwargs=self.remote_netcdf_kwargs)

        if len(table['paths'])>0:
            #Convert time axis to numbers and find the unique time axis:
            time_axis, time_axis_unique, date_axis_unique = unique_time_axis(date_axis,
                                                                             units, self.calendar,
                                                                             self.years, self.months)

            self.paths_ordering, paths_id_on_time_axis = reduce_paths_ordering(time_axis,
                                                                               time_axis_unique,
                                                                               self.paths_ordering,
                                                                               table)

            #Load data
            queryable_file_types_available = list(set(self.paths_ordering['file_type']).intersection(queryable_file_types))
            if len(queryable_file_types_available)>0:
                #Open the first file and use its metadata to populate container file:
                first_id = list(self.paths_ordering['file_type']).index(queryable_file_types_available[0])
                remote_data = remote_netcdf.remote_netCDF(self.paths_ordering['path'][first_id],
                                                        self.paths_ordering['file_type'][first_id],
                                                        semaphores=self.semaphores,
                                                        session=self.session,
                                                        **self.remote_netcdf_kwargs)
                time_dim,output = remote_data.safe_handling(netcdf_utils.find_time_dim_and_replicate_netcdf_file,output,time_var=self.time_var)
            else:
                remote_data = remote_netcdf.remote_netCDF(self.paths_ordering['path'][0],
                                                        self.paths_ordering['file_type'][0],
                                                        semaphores=self.semaphores,
                                                        session=self.session,
                                                        **self.remote_netcdf_kwargs)
                time_dim = self.time_var

            #Create time axis in ouptut:
            netcdf_utils.create_time_axis_date(output, date_axis_unique, units, self.calendar, time_dim=time_dim)

            self.create(output)
            if isinstance(var,list):
                for sub_var in var:
                    output = record_indices(self.paths_ordering,
                                               remote_data,output,sub_var,
                                               time_dim,time_axis,time_axis_unique,
                                               table,paths_id_on_time_axis,self.record_other_vars)
            else:
                output = record_indices(self.paths_ordering,
                                           remote_data,output,var,
                                           time_dim,time_axis,time_axis_unique,
                                           table,paths_id_on_time_axis,self.record_other_vars)
        return

def record_indices(paths_ordering,
                    remote_data,output,var,
                    time_dim,time_axis,time_axis_unique,
                    table,paths_id_on_time_axis,record_other_vars):
    #Create descriptive vars:
    #Must use compression, especially for ocean variables in curvilinear coordinates:
    remote_data.safe_handling(netcdf_utils.retrieve_variables_no_time,output,time_dim,zlib=True)

    #CREATE LOOK-UP TABLE:
    output_grp = output.groups['soft_links']
    indices_dim = 'indices'
    if not indices_dim in output_grp.dimensions:
        output_grp.createDimension(indices_dim,2)
    if not indices_dim in output_grp.variables.keys():
        output_grp.createVariable(indices_dim,np.str,(indices_dim,),chunksizes=(1,))
    indices = output_grp.variables[indices_dim]
    indices[0] = 'path'
    indices[1] = time_dim

    #Create soft links:
    paths_id_list = [path_id for path_id in paths_ordering['path_id'] ]

    #Replicate variable in main group:
    remote_data.safe_handling(netcdf_utils.replicate_netcdf_var,output,var,zlib=True)

    if var in output.variables.keys():
        var_out = output_grp.createVariable(var,np.int64,(time_dim,indices_dim),zlib=True)

        for time_id, time in enumerate(time_axis_unique):
            #For each time in time_axis_unique, pick path_id in paths_id_list. They
            #should all be the same. Pick the first one:
            paths_id_that_can_be_used = np.unique([path_id for path_id in paths_id_on_time_axis[time==time_axis]
                                            if path_id in paths_id_list])
            path_id_to_use = [path_id for path_id in paths_id_list
                                if path_id in paths_id_that_can_be_used][0]
            var_out[time_id,0] = path_id_to_use
            var_out[time_id,1] = table[indices_dim][np.logical_and(paths_id_on_time_axis==path_id_to_use,time==time_axis)][0]
        if np.ma.count_masked(var_out)>0:
            raise ValueError('Variable was not created properly. Must recreate')

    #Create support variables:
    if record_other_vars:
        previous_output_variables_list = output.variables.keys()
        #Replicate other vars:
        output = remote_data.safe_handling(netcdf_utils.replicate_netcdf_other_var,output,var,time_dim)
        output_variables_list = [ other_var for other_var in netcdf_utils.variables_list_with_time_dim(output,time_dim)
                                    if other_var!=var]
        for other_var in output_variables_list:
            if (not other_var in previous_output_variables_list):
                var_out = output_grp.createVariable(other_var,np.int64,(time_dim,indices_dim),zlib=True)
                #Create soft links:
                for time_id, time in enumerate(time_axis_unique):
                    #For each time in time_axis_unique, pick path_id in paths_id_list. They
                    #should all be the same. Pick the first one:
                    paths_id_that_can_be_used = np.unique([path_id for path_id in paths_id_on_time_axis[time==time_axis]
                                                    if path_id in paths_id_list])
                    path_id_to_use = [path_id for path_id in paths_id_list
                                        if path_id in paths_id_that_can_be_used][0]
                    var_out[time_id,0] = path_id_to_use
                    var_out[time_id,1] = table[indices_dim][np.logical_and(paths_id_on_time_axis==path_id_to_use,time==time_axis)][0]
                if np.ma.count_masked(var_out)>0:
                    raise ValueError('Variable was not created properly. Must recreate')
    output.sync()
    return output

def order_paths_by_preference(sorts_list,id_list,paths_list,file_type_list,data_node_list,check_dimensions,
                                semaphores=dict(),time_var='time',session=None,remote_netcdf_kwargs=dict()):
    #FIND ORDERING:
    paths_desc = []
    for id in sorts_list:
        paths_desc.append((id,np.int64))
    for id in id_list:
        paths_desc.append((id,'a255'))
    paths_ordering = np.empty((len(paths_list),), dtype=paths_desc)

    if check_dimensions:
        dimension_type_list = ['unqueryable',]

    for file_id, file in enumerate(paths_list):
        paths_ordering['path'][file_id] = file['path'].split('|')[0]
        #Convert path name to 'unique' integer using hash.
        #The integer will not really be unique but collisions
        #should be extremely rare for similar strings with only small variations.
        paths_ordering['path_id'][file_id] = hash(
                                                paths_ordering['path'][file_id]
                                                    )
        for unique_file_id in unique_file_id_list:
            paths_ordering[unique_file_id][file_id] = file['path'].split('|')[unique_file_id_list.index(unique_file_id)+1]
        paths_ordering['version'][file_id] = np.long(file['version'][1:])

        paths_ordering['file_type'][file_id] = file['file_type']
        paths_ordering['data_node'][file_id] = remote_netcdf.get_data_node(file['path'],paths_ordering['file_type'][file_id])
        
        if check_dimensions:
            #Dimensions types. Find the different dimensions types:
            if not paths_ordering['file_type'][file_id] in queryable_file_types:
                paths_ordering['dimension_type_id'][file_id] = dimension_type_list.index('unqueryable')
            else:
                remote_data = remote_netcdf.remote_netCDF(paths_ordering['path'][file_id],
                                                        paths_ordering['file_type'][file_id],
                                                        semaphores=semaphores,
                                                        session=session,
                                                        **remote_netcdf_kwargs)
                dimension_type = remote_data.safe_handling(netcdf_utils.find_dimension_type,time_var=time_var)
                if not dimension_type in dimension_type_list: dimension_type_list.append(dimension_type)
                paths_ordering['dimension_type_id'][file_id] = dimension_type_list.index(dimension_type)

    if check_dimensions:
        #Sort by increasing number. Later when we sort, we should get a uniform type:
        dimension_type_list_number = [ sum(paths_ordering['dimension_type_id']==dimension_type_id)
                                        for dimension_type_id,dimension_type in enumerate(dimension_type_list)]
        sort_by_number = np.argsort(dimension_type_list_number)[::-1]
        paths_ordering['dimension_type_id'] = sort_by_number[paths_ordering['dimension_type_id']]

    #Sort paths from most desired to least desired:
    #First order desiredness for least to most:
    data_node_order = copy.copy(data_node_list)[::-1]#list(np.unique(paths_ordering['data_node']))
    file_type_order = copy.copy(file_type_list)[::-1]#list(np.unique(paths_ordering['file_type']))
    for file_id, file in enumerate(paths_list):
        paths_ordering['data_node_id'][file_id] = data_node_order.index(paths_ordering['data_node'][file_id])
        paths_ordering['file_type_id'][file_id] = file_type_order.index(paths_ordering['file_type'][file_id])
    #'version' is implicitly from least to most

    #sort and reverse order to get from most to least:
    return np.sort(paths_ordering,order = sorts_list)[::-1]

def _recover_date(path,time_frequency,is_instant,calendar,semaphores=dict(),time_var='time',session=None,remote_netcdf_kwargs=dict()):
    file_type = path['file_type']
    path_name = str(path['path']).split('|')[0]
    remote_data = remote_netcdf.remote_netCDF(path_name,
                                            file_type,
                                            semaphores=semaphores,
                                            session=session,
                                            **remote_netcdf_kwargs)
    date_axis = remote_data.get_time(time_frequency=time_frequency,
                                    is_instant=is_instant,
                                    time_var=time_var,
                                    calendar=calendar)
    time_units = remote_data.get_time_units(calendar,time_var=time_var)
    table_desc = [
               ('paths','a255'),
               ('file_type','a255'),
               ('time_units','a255'),
               ('indices','int64')
               ] + [(unique_file_id,'a255') for unique_file_id in unique_file_id_list]
    if len(date_axis)>0:
        table = np.empty(date_axis.shape, dtype=table_desc)
        if len(date_axis)>0:
            table['paths'] = np.array([str(path_name) for item in date_axis])
            table['file_type'] = np.array([str(file_type) for item in date_axis])
            table['time_units'] = np.array([str(time_units) for item in date_axis])
            table['indices'] = range(0,len(date_axis))
            for unique_file_id in unique_file_id_list:
                table[unique_file_id] = np.array([str(path[unique_file_id]) for item in date_axis])
        return date_axis, table
    else:
        #No time axis, return empty arrays:
        return np.array([]), np.array([], dtype=table_desc)

def obtain_date_axis(paths_ordering,time_frequency,is_instant,calendar,semaphores=dict(),time_var='time',session=None,remote_netcdf_kwargs=dict()):
    #Retrieve time axes from queryable file types or reconstruct time axes from time stamp
    #from non-queryable file types.
    date_axis, table =  map(np.concatenate,
                    zip(*map(lambda x:_recover_date(x,time_frequency,
                                                      is_instant,
                                                      calendar,
                                                      semaphores=semaphores,
                                                      time_var=time_var,
                                                      session=session,
                                                      remote_netcdf_kwargs=remote_netcdf_kwargs),np.nditer(paths_ordering))))
    if len(date_axis)>0:
        #If all files have the same time units, use this one. Otherwise, create a new one:
        unique_date_units = _drop_none(np.unique(table['time_units']))
        if len(unique_date_units) == 1:
            units = unique_date_units[0]
        elif len(unique_date_units) > 1:
            #Use units with earliest since date:
            units_id = np.argsort(map(lambda x: x.split('since')[-1], unique_date_units))[0]
            units = unique_date_units[units_id]
        else:
            units = 'days since ' + str(date_axis[0])
    return date_axis, table, units

def _drop_none(arr):
    return arr[arr != np.array(None)]

def _recover_calendar(path,semaphores=dict(),time_var='time',session=None,remote_netcdf_kwargs=dict()):
    file_type = path['file_type']
    path_name = str(path['path']).split('|')[0]
    remote_data = remote_netcdf.remote_netCDF(path_name,
                                            file_type,
                                            semaphores=semaphores,
                                            session=session,
                                            **remote_netcdf_kwargs)
    calendar = remote_data.get_calendar(time_var=time_var)
    return calendar, file_type 

def obtain_unique_calendar(paths_ordering,semaphores=dict(),time_var='time',session=None,remote_netcdf_kwargs=dict()):
    calendar_list,file_type_list=zip(*map(lambda x:_recover_calendar(x,semaphores=semaphores,
                                                                        time_var=time_var,
                                                                        session=session,
                                                                        remote_netcdf_kwargs=remote_netcdf_kwargs),np.nditer(paths_ordering)))
    #Find the calendars found from queryable file types:
    calendars = set([item[0] for item in zip(calendar_list,file_type_list) if item[1] in queryable_file_types])
    if len(calendars)==1:
        return calendars.pop()
    return calendar_list[0]

def reduce_paths_ordering(time_axis,time_axis_unique,paths_ordering,table):
    #CREATE LOOK-UP TABLE:
    paths_indices_on_time_axis = np.empty(time_axis.shape,dtype=np.int64)
    paths_id_on_time_axis = np.empty(time_axis.shape,dtype=np.int64)

    paths_list = list(paths_ordering['path'])
    paths_id_list = list(paths_ordering['path_id'])
    #for path_id, path in zip(paths_id_list,paths_list):
    for path_index, (path_id, path) in enumerate(zip(paths_id_list,paths_list)):
        #find in table the path and assign path_index to it:
        paths_indices_on_time_axis[path==table['paths']] = path_index
        paths_id_on_time_axis[path==table['paths']] = path_id

    #Remove paths that are not necessary over the requested time range:
    #First, list the paths_id used:
    #Pick the lowest path index so that we follow the paths_ordering:
    useful_paths_id_list_unique = list(
                                np.unique(
                                    [paths_id_list[np.min(paths_indices_on_time_axis[time==time_axis])] 
                                        for time_id, time in enumerate(time_axis_unique)]))

    #Second, list the path_names corresponding to these paths_id:
    #useful_file_name_list=[useful_path_id.split('/')[-1] for useful_path_id in 
    #                        [path for path_id, path in zip(paths_id_list,paths_list)
    #                                if path_id in useful_paths_id_list_unique] ]
    useful_file_name_list_unique = [paths_list[paths_id_list.index(path_id)].split('/')[-1]
                                        for path_id in useful_paths_id_list_unique]

    #Find the indices to keep:
    useful_file_id_list = [file_id for file_id, file in enumerate(paths_ordering)
                            if paths_ordering['path_id'][file_id] in useful_paths_id_list_unique]
                            
    #Finally, check if some equivalent indices are worth keeping:
    for file_id, file in enumerate(paths_ordering):
        if not paths_ordering['path_id'][file_id] in useful_paths_id_list_unique:
            #This file was not kept but it might be the same data, in which
            #case we would like to keep it.
            #Find the file name (remove path):
            file_name = paths_ordering['path'][file_id].split('/')[-1]
            if file_name in useful_file_name_list_unique:
                #If the file name is useful, find its path_id: 
                equivalent_path_id = useful_paths_id_list_unique[useful_file_name_list_unique.index(file_name)]
                    
                #Use this to find its file_id:
                equivalent_file_id = list(paths_ordering['path_id']).index(equivalent_path_id)
                #Then check if the checksum are the same. If yes, keep the file!
                if paths_ordering['checksum'][file_id]==paths_ordering['checksum'][equivalent_file_id]:
                    useful_file_id_list.append(file_id)
            
    #Sort paths_ordering:
    if len(useful_file_id_list)>0:
        return paths_ordering[np.sort(useful_file_id_list)],paths_id_on_time_axis
    else:   
        return paths_ordering,paths_id_on_time_axis

def unique_time_axis(date_axis,units,calendar,years,months):
    time_axis = netcdf_utils.get_time_axis_relative(date_axis,units,calendar)
    time_axis_unique = np.unique(time_axis)

    date_axis_unique = netcdf_utils.get_date_axis_relative(time_axis_unique,units,calendar)

    #Include a filter on years and months: 
    time_desc = {}
    if years!=None:
        if years[0]<10:
            #This is important for piControl
            temp_years = list(np.array(years)+np.min([date.year for date in date_axis_unique]))
            #min_year = np.min([date.year for date in date_axis_unique])
        else:
            temp_years = years
        if months!=None:
            valid_times = np.array([True if (date.year in temp_years and 
                                     date.month in months) else False for date in  date_axis_unique])
        else:
            valid_times = np.array([True if date.year in temp_years else False for date in  date_axis_unique])
    else:
        if months!=None:
            valid_times = np.array([True if date.month in months else False for date in  date_axis_unique])
        else:
            valid_times = np.array([True for date in  date_axis_unique])
        
    #self.time_axis_unique = time_axis_unique[valid_times]
    #self.date_axis_unique = date_axis_unique[valid_times]
    #self.time_axis = time_axis
    return time_axis, time_axis_unique[valid_times], date_axis_unique[valid_times]



#External:
import os

# Define a context manager to suppress stdout and stderr.
#http://stackoverflow.com/questions/11130156/suppress-stdout-stderr-print-from-python-functions
class suppress_stdout_stderr(object):
    '''
    A context manager for doing a "deep suppression" of stdout and stderr in 
    Python, i.e. will suppress all print, even if the print originates in a 
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).      

    '''
    def __init__(self):
        # Open a pair of null files
        self.null_fds =  [os.open(os.devnull,os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = (os.dup(1), os.dup(2))
        #print(self.null_fds,self.save_fds)
        return

    def __enter__(self):

        # Assign the null pointers to stdout and stderr.
        for id in range(2):
            os.dup2(self.null_fds[id],id+1)
        return

    def __exit__(self, *_):
        for id in range(2):
            # Re-assign the real stdout/stderr back to (1) and (2)
            os.dup2(self.save_fds[id],id+1)
        return

    def close(self):
        for id in range(2):
            # Close the null files
            os.close(self.null_fds[id])
        #Close the duplicates:
        #Very important otherwise "too many files open"
        map(os.close,self.save_fds)
        return

class dummy_semaphore:
    def acquire(self):
        return 
    def release(self):
        return




def onlineca_get_trustroots_wget():
    return """
#!/bin/bash
#
# Client script for web service interface to SLCS get-trustroots based on 
# curl and base64 commands.  Get trust roots retrieves the CA certificate 
# issuer(s) of the SLCS's SSL certificate
#
# @author P J Kershaw 07/06/2010
#
# @copyright: (C) 2010 STFC
#
# @license: BSD - See top-level LICENCE file for licence details
#
# $Id$
cmdname=$(basename $0)
cmdline_opt=`getopt hU:bc: $*`

usage="Usage: $cmdname [-h][-U get trust roots URI][-b][-c CA directory]\n
\n
   Options\n
       -h\t\t\tDisplays usage and quits.\n
       -U <uri>\t\tShort-Lived Credential service URI\n
       -b\t\t\tbootstrap trust in the Short-Lived Credential Service\n
       -c <directory path>\tDirectory to store the trusted CA (Certificate Authority) certificates.\n
       \t\t\tDefaults to ${HOME}/.globus/certificates or\n
       \t\t\t/etc/grid-security/certificates if running as root.\n
"

if [ $? != 0 ] ; then
    echo -e $usage >&2 ;
    exit 1 ;
fi

set -- $cmdline_opt

while true ; do
    case "$1" in
        -h) echo -e $usage ; exit 0 ;;
        -U) uri=$2 ; shift 2 ;;
        -b) bootstrap=1 ; shift 1 ;;
        -c) cadir=$2 ; shift 2 ;;
         --) shift ; break ;;
        *) echo "Error parsing command line" ; exit 1 ;;
    esac
done

if [ -z $uri ]; then
    echo -e Give the URI for the Short-Lived Credential service get trust roots request;
    echo -e $usage >&2 ;
    exit 1;
fi

# Set-up destination trust root directory
if [ $cadir ]; then
    if [ ! -d $cadir ]; then
        mkdir -p $cadir
    fi
    
elif [ ${X509_CERT_DIR} ]; then
    cadir=${X509_CERT_DIR}
    
elif [ "$LOGNAME" = "root" ]; then
    cadir=/etc/grid-security/certificates
    
    # Check path exists and if not make it
    if [ ! -d "/etc/grid-security" ]; then
        mkdir /etc/grid-security
    fi
       
    if [ ! -d "/etc/grid-security/certificates" ]; then
        mkdir /etc/grid-security/certificates
    fi
else
    cadir=${HOME}/.globus/certificates
    
    # Check path exists and if not make it
    if [ ! -d "${HOME}/.globus" ]; then
        mkdir ${HOME}/.globus
    fi
    
    if [ ! -d "${HOME}/.globus/certificates" ]; then
        mkdir ${HOME}/.globus/certificates
    fi
fi

# Set peer authentication based on bootstrap command line setting
if [ -z $bootstrap ]; then 
    ca_arg="--ca-directory $cadir"
else
    #echo Bootstrapping Short-Lived Credential Service root of trust.
    ca_arg="--no-check-certificate"
fi

# Make a temporary file for error output
error_output_filepath="/tmp/$UID-$RANDOM.csr"

# Post request to Short-Lived Credential Service
response=$(wget $uri  --secure-protocol TLSv1 $ca_arg -t 1 -O - 2> $error_output_filepath)

# Extract error output and clean up
error_output=$(cat $error_output_filepath)
rm -f $error_output_filepath

# Pull out the response code from the error output
wget_statcode_line="awaiting response..."
responsecode=$(echo "$error_output"|grep "$wget_statcode_line"|awk '{print $6}')
if [ "$responsecode" != "200" ]; then
    echo "Get trust roots failed"
    echo "$error_output" >&2
    exit 1
fi

# Process response
entries=$(echo $response|awk '{print $0}')
for i in $entries; do
    filename=${i%%=*}
    filecontent="$(echo ${i#*=}|awk '{for(i=1;i<length;i+=65) print substr($0,i,65)}'|openssl enc -d -base64)"
    echo "$filecontent" > $cadir/$filename
done

#echo Trust roots have been installed in $cadir.
"""



from __future__ import division, absolute_import, print_function

#External:
import netCDF4
import numpy as np
import itertools
import scipy.interpolate as interpolate
import scipy.spatial as spatial
import copy
#import spherical_geometry.great_circle_arc as great_circle_arc
import pandas

#Internal:
from .. import netcdf_utils

default_box=[0.0,360.0,-90.0,90.0]
def subset(input_file,output_file,lonlatbox=default_box,lat_var='lat',lon_var='lon',output_vertices=False):
    """
    Function to subset a hierarchical netcdf file. Its latitude and longitude
    should follow the CMIP5 conventions.
    """
    #Modify lonlatbox to handle periodic longitudes
    mod_lonlatbox=np.array(copy.copy(lonlatbox))
    if np.diff(np.mod(lonlatbox[:2],360))==0:
        if np.diff(lonlatbox[:2])>0:
            mod_lonlatbox[1]-=1e-6
        elif np.diff(lonlatbox[:2])<0:
            mod_lonlatbox[1]+=1e-6
    optimal_slice = (lambda x: get_optimal_slices(x,mod_lonlatbox,lat_var,lon_var,output_vertices))
    with netCDF4.Dataset(input_file,'r') as dataset:
        with netCDF4.Dataset(output_file,'w') as output:
            if output_vertices:
                transform = (lambda x,y,z: get_and_write_vertices(x,y,lat_var,lon_var,z))
                netcdf_utils.replicate_full_netcdf_recursive(dataset,output,transform=transform,slices=optimal_slice,check_empty=True)
            else:
                netcdf_utils.replicate_full_netcdf_recursive(dataset,output,slices=optimal_slice,check_empty=True)
    return

def get_optimal_slices(data,lonlatbox,lat_var,lon_var,output_vertices):
    if set([lat_var,lon_var]).issubset(data.variables.keys()):
        lat=data.variables[lat_var][:]
        lon=np.mod(data.variables[lon_var][:],360.0)
        if output_vertices or check_basic_consistency(data,lat_var,lon_var):
            lat_vertices, lon_vertices=get_vertices(data,lat_var,lon_var)
            region_mask=get_region_mask(lat_vertices,lon_vertices,lonlatbox)
            if ( set([lat_var+'_bnds',lon_var+'_bnds']).issubset(data.variables.keys())
                and not ( data.variables[lat_var+'_bnds'].shape==data.variables[lat_var].shape+(4,) and
                          data.variables[lon_var+'_bnds'].shape==data.variables[lon_var].shape+(4,))):
                dimensions=(lat_var,lon_var)
            else:
                dimensions=data.variables[lat_var].dimensions
        else:
            if len(lat.shape) == 1 and len(lon.shape) == 1:
                # Broadcast:
                LON, LAT = np.meshgrid(lon, lat)
                region_mask=get_region_mask(LAT[...,np.newaxis],LON[...,np.newaxis],lonlatbox)
            else:
                region_mask=get_region_mask(lat[...,np.newaxis],lon[...,np.newaxis],lonlatbox)
            if ( data.variables[lat_var].dimensions==(lat_var,) and
               data.variables[lon_var].dimensions==(lon_var,) ):
               dimensions=(lat_var,lon_var)
            else:
               dimensions=data.variables[lat_var].dimensions

        return {dimensions[id]:
                       np.arange(region_mask.shape[id])[np.sum(region_mask,axis=1-id)>0] for id in [0,1]}
    else:
        return dict()

def get_region_mask(lat,lon,lonlatbox):
    """
    lat and lon must have an extra trailing dimension that can correspond to
    a vertices dimension
    """
    if np.diff(np.mod(lonlatbox[:2],360))<0:
        lon_region_mask=np.logical_not(np.logical_and(
                                   np.min(np.mod(lon,360),axis=-1)>=np.mod(lonlatbox[1],360),
                                   np.max(np.mod(lon,360),axis=-1)<=np.mod(lonlatbox[0],360)))
    else:
        lon_region_mask=np.logical_and(
                                   np.min(np.mod(lon,360),axis=-1)>=np.mod(lonlatbox[0],360),
                                   np.max(np.mod(lon,360),axis=-1)<=np.mod(lonlatbox[1],360))
    lat_region_mask=np.logical_and(
                                   np.min(lat,axis=-1)>=lonlatbox[2],
                                   np.max(lat,axis=-1)<=lonlatbox[3])
    return np.logical_and(lon_region_mask,lat_region_mask)


def get_and_write_vertices(data,output,lat_var,lon_var,comp_slices):
    if ( not lat_var+'_vertices' in output.variables.keys() or
         not lon_var+'_vertices' in output.variables.keys() ):
        lat_vertices,lon_vertices=get_vertices(data,lat_var,lon_var)
        record_vertices(data,output,lat_var,lat_vertices,comp_slices)
        record_vertices(data,output,lon_var,lon_vertices,comp_slices)
    return output

def record_vertices(data,output,var,vertices,comp_slices):
    dim='nv'
    if not dim in output.dimensions.keys():
        output.createDimension('nv',size=4)
    out_dims=data.variables[var].dimensions+('nv',)
    getitem_tuple=tuple([comp_slices[var_dim] if var_dim in comp_slices.keys()
                                                else slice(None,None,1) for var_dim in out_dims])
    temp=output.createVariable(var+'_vertices','f',out_dims)
    sliced_vertices=vertices
    for axis_id,indexing in enumerate(getitem_tuple):
        if isinstance(indexing,slice):
            sliced_vertices=np.ma.take(sliced_vertices,np.arange(sliced_vertices.shape[axis_id])[indexing],axis=axis_id)
        else:
            sliced_vertices=np.ma.take(sliced_vertices,indexing,axis=axis_id)
    temp[:]=sliced_vertices
    return

def get_vertices(data,lat_var,lon_var):
    if not set([lat_var+'_vertices',lon_var+'_vertices']).issubset(data.variables.keys()):
        if set([lat_var+'_bnds',lon_var+'_bnds']).issubset(data.variables.keys()):
            if ( data.variables[lat_var+'_bnds'].shape==data.variables[lat_var].shape+(4,) and
                 data.variables[lon_var+'_bnds'].shape==data.variables[lon_var].shape+(4,)):
                #lat_bnds and lon_bnds are in fact lat and lon vertices: 
                lat_vertices=data.variables[lat_var+'_bnds'][:]
                lon_vertices=data.variables[lon_var+'_bnds'][:]
            else:
                lat_vertices,lon_vertices=get_vertices_from_bnds(data.variables[lat_var+'_bnds'][:],
                                                                 np.mod(data.variables[lon_var+'_bnds'][:],360))
        elif set(['rlat_bnds','rlon_bnds']).issubset(data.variables.keys()):
            lat_vertices,lon_vertices=get_spherical_vertices_from_rotated_bnds(data.variables['rlat'][:],
                                                                               data.variables['rlon'][:],
                                                                               data.variables[lat_var][:],
                                                                               np.mod(data.variables[lon_var][:],360),
                                                                               data.variables['rlat_bnds'][:],
                                                                               data.variables['rlon_bnds'][:])
        else:
            lat_vertices,lon_vertices=get_vertices_voronoi(data.variables[lat_var][:],data.variables[lon_var][:])
    else:
        lat_vertices=data.variables[lat_var+'_vertices'][:]
        lon_vertices=np.mod(data.variables[lon_var+'_vertices'][:],360)
    return lat_vertices, lon_vertices

def get_vertices_voronoi(lat,lon,do_not_simplify_edge_number=4):
    """
    A general method to obtain grid vertices based on voronoi diagrams.
    """
    r=1.0 
    if len(lat.shape)==1 and len(lon.shape)==1:
        LON, LAT = np.meshgrid(lon,lat)
    elif lat.shape==lon.shape:
        LON, LAT = lon, lat
    else:
        raise InputError('latitude variable and longitute variable must either both be vectors or\
                          have the same shape')
    shape=LON.shape
    x, y, z = np.vectorize(sc_to_rc)(r,LAT,LON)

    points=np.concatenate((x.ravel()[:,np.newaxis],
                           y.ravel()[:,np.newaxis],
                           z.ravel()[:,np.newaxis]),axis=1)
    mask=np.logical_or.reduce(np.logical_not(np.ma.getmaskarray(points)),1)
    voronoi_diag=spatial.SphericalVoronoi(np.ma.filled(points[mask,:],fill_value=np.nan),radius=r)

    #create edges dataframe:
    df_regions=regions_dataframe(lat.ravel(),lon.ravel())
    df_vertices=vertices_dataframe(voronoi_diag.vertices)

    #sort vertices in counterclockwise direction:
    voronoi_diag.sort_vertices_of_regions()
    #Creates edges dataframe:
    df_edges=pd.concat(map(lambda x: region_edges_dataframe(*x), enumerate(voronoi_diag.regions)))

    edges_df=pd.concat(map(lambda x: get_region_edges(*x),enumerate(voronoi_diag.regions)))
    #label edges:
    edges_df['edge_id']=label_edges(edges_df)

    region_edges=edges_df.groupby('region').size()
    regions_to_consider=regions_edges['region'][region_edges['size']>do_not_simplify_edge_number,:]

    
    simplified_vertices_of_regions=map(simplify_to_four_spherical_vertices_recursive,vertices_of_regions)
    lat_vertices[mask,:],lon_vertices[mask,:] = map(np.concatenate,
                                                    zip(*simplified_vertices_of_regions))
    return map(lambda x: np.reshape(np.ma.fix_invalid(x),shape+(4,)),[lat_vertices,lon_vertices])

def regions_dataframe(lat,lon):
    df=pd.DataFrame()
    df['lat']=lat
    df['lon']=lon
    return df

def vertices_dataframe(vertices):
    df = pd.DataFrame()
    df['x']=zip(*vertices)[0]
    df['y']=zip(*vertices)[1]
    df['z']=zip(*vertices)[2]
    return df

def region_edges_dataframe(region_id,vx_ids):
    df=pd.DataFrame()
    df['A']=vx_ids
    df['B'][:-1]=vx_ids[1:]
    df['B'][-1]=vx_ids[0]
    df['region']=region_id
    #df['reversed']=False
    #df.loc[(df['A']
    return df

def get_region_vertices(vertices,region_indices):
    return np.take(vertices,region_indices,axis=0)

#def simplify_to_four_spherical_vertices_recursive(sorted_vertices):
#    if sorted_vertices.shape[0]==3:
#        return convert_to_lat_lon(np.concatenate((sorted_vertices,np.array([np.nan,np.nan,np.nan])[np.newaxis,:]),axis=0))
#    elif sorted_vertices.shape[0]==4:
#        return convert_to_lat_lon(sorted_vertices)
#    else:
#        min_id=find_minimum_arc_id(sorted_vertices)
#        midpoint=arc_midpoint(sorted_vertices[min_id,:],sorted_vertices[np.mod(min_id+1,sorted_vertices.shape[0]),:])
#        fewer_sorted_vertices=np.empty((sorted_vertices.shape[0]-1,3))
#        if min_id==sorted_vertices.shape[0]:
#            fewer_sorted_vertices[:-1,:]=sorted_vertices[1:min_id,:]
#            fewer_sorted_vertices[-1,:]=midpoint
#        else:
#            fewer_sorted_vertices[:min_id,:]=sorted_vertices[:min_id,:]
#            fewer_sorted_vertices[min_id,:]=midpoint
#            fewer_sorted_vertices[min_id+1:,:]=sorted_vertices[min_id+2:,:]
#        return simplify_to_four_spherical_vertices_recursive(fewer_sorted_vertices)
#
#def find_minimum_arc_id(sorted_vertices):
#    return np.argmin(map(lambda x: arc_length(sorted_vertices[x,:],
#                                              sorted_vertices[np.mod(x+1,sorted_vertices.shape[0]),:]),
#                                    range(len(sorted_vertices)-1)))
#
#def arc_length(a,b):
#    if np.allclose(a,b):
#        return 0.0
#    else:
#        return great_circle_arc.length(a,b)
#
#def arc_midpoint(a,b):
#    if np.allclose(a,b):
#        midpoint=0.5*(a+b)
#        midpoint*=np.sqrt((a**2).sum())/np.sqrt((midpoint**2).sum())
#        return midpoint
#    else:
#        return great_circle_arc.midpoint(a,b)

def convert_to_lat_lon(sorted_vertices):
    return map(np.transpose,np.split(np.apply_along_axis(rc_to_sc_vec,1,sorted_vertices)[:,1:],2,axis=1))

def get_vertices_from_bnds(lat_bnds,lon_bnds):
    #Create 4 vertices:
    return np.broadcast_arrays(np.append(lat_bnds[:,np.newaxis,:],lat_bnds[:,np.newaxis,:],axis=-1),
                               np.insert(lon_bnds[np.newaxis,:,:],[0,1],lon_bnds[np.newaxis,:,:],axis=-1))


def sort_vertices_counterclockwise_array(lat_vertices, lon_vertices):
    struct=np.empty(lat_vertices.shape,dtype=[('lat_vertices',lat_vertices.dtype),
                                              ('lon_vertices',lat_vertices.dtype)])
    struct['lat_vertices']=np.ma.filled(lat_vertices,fill_value=np.nan)
    struct['lon_vertices']=np.ma.filled(lon_vertices,fill_value=np.nan)
    out_struct=np.apply_along_axis(sort_vertices_counterclockwise_struct,-1,struct)
    return np.ma.fix_invalid(out_struct['lat_vertices']), np.ma.fix_invalid(out_struct['lon_vertices'])
    #return (np.ma.masked_where(mask,out_struct['lat_vertices']),
    #        np.ma.masked_where(mask,out_struct['lon_vertices']))

def sort_vertices_counterclockwise_struct(struct):
    out_struct=np.empty_like(struct)
    out_struct['lat_vertices'],out_struct['lon_vertices']=map(lambda x: np.ma.filled(x,fill_value=np.nan),
                                                                sort_vertices_counterclockwise(np.ma.fix_invalid(struct['lat_vertices']),
                                                                                               np.ma.fix_invalid(struct['lon_vertices'])))
    return out_struct

def sort_vertices_counterclockwise(lat_vertices,lon_vertices):
    '''
    Ensure that vertices are listed in a counter-clockwise fashion
    '''
    vec=np.ma.concatenate(np.vectorize(sc_to_rc)(1.0,lat_vertices[:,np.newaxis],lon_vertices[:,np.newaxis]),axis=1)
    vec_c=np.ma.mean(vec,axis=0)
    vec-=vec_c[np.newaxis,:]

    cross=np.zeros((4,4))
    for i in range(cross.shape[0]):
        for j in range(cross.shape[1]):
            cross[i,j]=np.ma.dot(vec_c,np.cross(vec[i,:],vec[j,:]))

    id0=np.argmax(np.mod(lon_vertices,360))
    for id1 in range(4):
        for id2 in range(4):
            for id3 in range(4):
                if (len(set([id0,id1,id2,id3]))==4 and
                    cross[id0,id1]>0.0 and
                    cross[id1,id2]>0.0 and
                    cross[id2,id3]>0.0 and
                    cross[id3,id0]>0.0):
                    id_list=np.array([id0,id1,id2,id3])
                    return lat_vertices[id_list],lon_vertices[id_list]
    return lat_vertices, lon_vertices

def rc_to_sc_vec(point):
    return np.array(rc_to_sc(*point))

def rc_to_sc(x,y,z):
    '''
    Spherical coordinates to rectangular coordiantes.
    '''
    if np.any(np.isnan([x,y,z])):
        return np.nan, np.nan,np.nan

    r=np.sqrt(x**2+y**2+z**2)
    if z==0.0:
        lat=0.0
    else:
        lat=(0.5*np.pi-np.arccos(z/r))*180.0/np.pi
    if x==0.0:
        lon=90.0
    else:
        lon=np.arctan(y/x)*180.0/np.pi
    return r, lat, lon

def sc_to_rc(r,lat,lon):
    '''
    Spherical coordinates to rectangular coordiantes.
    '''
    x=r*np.sin(0.5*np.pi-lat/180.0*np.pi)*np.cos(lon/180.0*np.pi)
    y=r*np.sin(0.5*np.pi-lat/180.0*np.pi)*np.sin(lon/180.0*np.pi)
    z=r*np.cos(0.5*np.pi-lat/180.0*np.pi)
    return x,y,z

def get_spherical_vertices_from_rotated_bnds(rlat,rlon,lat,lon,rlat_bnds,rlon_bnds):
    '''
    It is assumed that input longitudes (but not rotated longitudes) are 0 to 360 degrees
    '''
    rlat_vertices,rlon_vertices=get_vertices_from_bnds(rlat_bnds,rlon_bnds)
    rescale=np.pi/180.0
    rlon_bnds=fix_lon_bnds(rlon_bnds)
    rlon_offset=np.min(rlon_bnds)
    rlat_valid_points=np.logical_and(rlat_vertices<np.max(rlat),rlat_vertices>np.min(rlat))
    lat_vertices=np.ma.empty_like(rlat_vertices)
    lat_vertices[np.logical_not(rlat_valid_points)]=np.ma.masked
    lat_vertices[rlat_valid_points]=spherical_interp((rlat+90.0)*rescale,(rlon-rlon_offset)*rescale,lat,(rlat_vertices[rlat_valid_points]+90.0)*rescale,
                                                                                                   (rlon_vertices[rlat_valid_points]-rlon_offset)*rescale)
    rlon_valid_points=np.logical_and(rlon_vertices<np.max(rlon),rlon_vertices>np.min(rlon))
    lon_vertices=np.ma.empty_like(rlon_vertices)
    lon_vertices[np.logical_not(rlat_valid_points)]=np.ma.masked
    lon_vertices[rlat_valid_points]=spherical_interp((rlat+90.0)*rescale,(rlon-rlon_offset)*rescale,lon,(rlat_vertices[rlat_valid_points]+90.0)*rescale,
                                                                                                   (rlon_vertices[rlat_valid_points]-rlon_offset)*rescale)
    lon_vertices_mod=np.ma.empty_like(rlon_vertices)
    lon_vertices_mod[np.logical_not(rlat_valid_points)]=np.ma.masked
    lon_vertices_mod[rlat_valid_points]=np.mod(spherical_interp((rlat+90.0)*rescale,(rlon-rlon_offset)*rescale,np.mod(lon-180.0,360),(rlat_vertices[rlat_valid_points]+90.0)*rescale,
                                                                                                   (rlon_vertices[rlat_valid_points]-rlon_offset)*rescale)+180.0,360)
                   
    lon_vertices[lon<90.0,:]=lon_vertices_mod[lon<90.0,:]
    lon_vertices[lon>270.0,:]=lon_vertices_mod[lon>270.0,:]

    return lat_vertices, lon_vertices

def spherical_interp(rlat,rlon,arr,rlat_vertices,rlon_vertices):
    interpolants_simple=interpolate.RectSphereBivariateSpline(rlat,rlon,arr)
    interpolants=(lambda x: interpolants_simple.ev(*x))
    N=100
    return np.concatenate(map(interpolants,zip(np.array_split(rlat_vertices,N),
                                               np.array_split(rlon_vertices,N))),axis=0)

def fix_lon_bnds(lon_bnds):
    range=lon_bnds.max()-lon_bnds.min()
    if range<360.0:
        diff=360.0-range
        max_diff=np.max(np.diff(lon_bnds,axis=1))
        if (diff-0.1<=max_diff):
            lon_bnds[0,0]-=diff*0.5
            lon_bnds[-1,1]+=diff*0.5
    return lon_bnds

def check_basic_consistency(data,lat_var,lon_var):
    coords=[(lat_var,lon_var),('rlat','rlon')]
    bnds=['vertices','bnds']
    has_coordinates_bnds=np.any([ set(coordinates_bnds).issubset(data.variables.keys())
                                for coordinates_bnds in 
                                        itertools.chain.from_iterable([[ [single_coord+'_'+bnd 
                                                                        for single_coord in coord] 
                                                                    for coord in coords] 
                                                                        for bnd in bnds])] )
    if not has_coordinates_bnds:
        return False
    return True



# This Python file uses the following encoding: utf-8


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    import os
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

from setuptools import setup, find_packages

package_name='netcdf4_soft_links'
setup(
       name = package_name,
       version = "0.7.8.1",
       packages = find_packages(),
#
#        # metadata for upload to PyPI
        author = "F. B. Laliberte P. J. Kushner",
        author_email = "frederic.laliberte@utoronto.ca",
        description = "Tools to create soft links to netcdf4 files.",
        license = "BSD",
        keywords = "atmosphere climate",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "Natural Language :: English",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Fortran",
            "Topic :: Scientific/Engineering :: Mathematics"
        ],
        long_description=read('README.rst'),
        install_requires = ['numpy',
                            'h5py',
                            'h5netcdf>=0.3',
                            'netCDF4',
                            'netcdf4_pydap>=0.2.0.1',
                            'requests>=1.1.0',
                            'requests_cache'
                            ],
        zip_safe=False,
        # other arguments here...
        entry_points = {
                  'console_scripts': [
                           'nc4sl='+package_name+'.core:main'
                                     ],
                       }
    )
